/*
 * Copyright (c) 2001 Fabrice Bellard
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/**
 * @file libavcodec encoding video API usage example
 * @example encode_video.c
 *
 * Generate synthetic video data and encode it to an output file.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>



#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libavcodec/avcodec.h>

#include <libavutil/opt.h>
#include <libavutil/imgutils.h>

#include <libswscale/swscale.h>

int compare_strings(const void *a, const void *b) {
  return strcmp(*(const char**)a, *(const char**)b);
}

void fatal(const char *fmt, ...) {
    va_list args;
    char buf[1024];  // Buffer to hold the formatted message
    va_start(args, fmt);
    vsnprintf(buf, sizeof(buf), fmt, args);  // Format the message into the buffer
    va_end(args);
    perror(buf);  // Pass the buffer as the argument to perror
    exit(EXIT_FAILURE);
}


unsigned char *read_file(char *filename, size_t *len_buffer) {
  FILE *file = fopen(filename, "rb");
  if (file == NULL) {
    fatal("Failed to open input file");
  }

  // Get the size of the input file
  fseek(file, 0, SEEK_END);
  long filesize = ftell(file);
  fseek(file, 0, SEEK_SET);

  // Allocate a buffer for the file data
  unsigned char *buffer = (unsigned char *) malloc(filesize);
  if (buffer == NULL) {
    fatal("Failed to allocate memory for file data");
  }
  *len_buffer = filesize;

  // Read the file data into the buffer
  size_t bytes_read = fread(buffer, 1, filesize, file);
  if (bytes_read != filesize) {
    fatal("Failed to read file data");
  }
  if (ferror(file)) {
    fatal("Failed to read file");
  }

  // Close the input file
  fclose(file);

  return buffer;
}

typedef struct {
  unsigned char *data;
  size_t size;
  int64_t pts;
  int has_data;
} Mypacket;

#define MAX_PACKETS 10

Mypacket buffer[MAX_PACKETS];

void add_packet(AVPacket *pkt) {
  for (int i = 0; i < MAX_PACKETS; i++) {
    if (!buffer[i].has_data) {
      buffer[i].data = (unsigned char *) malloc(pkt->size);
      memcpy(buffer[i].data, pkt->data, pkt->size);
      buffer[i].size = pkt->size;
      buffer[i].pts = pkt->pts;
      buffer[i].has_data = 1;
      printf("added packet pkt %ld at position %d\n", pkt->pts, i);
      return;
    }
    fatal("packet buffer not large enough");
  }
}

void flush_packets(int64_t *out_pts, FILE *file) {
  puts("flushing packets");
  for (int i = 0; i < MAX_PACKETS; i++) {
    if (buffer[i].has_data && buffer[i].pts == *out_pts) {
      printf("flushing at position %d, pts %ld\n", i, buffer[i].pts);
      printf("wr %ld\n", *out_pts);
      fwrite(buffer[i].data, 1, buffer[i].size, file);
      free(buffer[i].data);
      buffer[i].has_data = 0;
      *out_pts += 1;
      flush_packets(out_pts, file);
      return;
    }
  }
}

void write_packet(AVPacket *pkt, int64_t *out_pts, FILE *file) {
  if (pkt->pts == *out_pts) {
    printf("wr %ld\n", *out_pts);
    fwrite(pkt->data, 1, pkt->size, file);
    *out_pts += 1;
    flush_packets(out_pts, file);
  }
  else {
    add_packet(pkt);
  }
}

static void encode(AVCodecContext *enc_ctx, AVFrame *frame, AVPacket *pkt,
                   FILE *outfile)
{
    int ret;

    static int64_t out_pts = 0;

    /* send the frame to the encoder */
    if (frame)
        printf("Send frame %3"PRId64"\n", frame->pts);

    ret = avcodec_send_frame(enc_ctx, frame);
    if (ret < 0) {
        fprintf(stderr, "Error sending a frame for encoding\n");
        exit(1);
    }

    while (ret >= 0) {
        ret = avcodec_receive_packet(enc_ctx, pkt);
        if (ret == AVERROR(EAGAIN) || ret == AVERROR_EOF)
            return;
        else if (ret < 0) {
            fprintf(stderr, "Error during encoding\n");
            exit(1);
        }

        /* printf("Write packet %3"PRId64" (size=%5d)\n", pkt->pts, pkt->size); */
	/* printf("pkt: %p\n", (void *) pkt); */
	/* printf("pkt->data: %p\n", (void *) pkt->data); */
        //fwrite(pkt->data, 1, pkt->size, outfile);
	write_packet(pkt, &out_pts, outfile);
        av_packet_unref(pkt);
    }
}

int main(int argc, char **argv)
{
    const char *filename, *codec_name;
    const AVCodec *codec;
    AVCodecContext *c= NULL;
    int i, ret, x, y;
    FILE *f;
    AVFrame *frame;
    AVPacket *pkt;
    uint8_t endcode[] = { 0, 0, 1, 0xb7 };

    
    if (argc <= 2) {
        fprintf(stderr, "Usage: %s <output file> <codec name>\n", argv[0]);
        exit(0);
    }
    filename = argv[1];
    codec_name = argv[2];


  // Open the input directory
  DIR *dir = opendir("input");
  if (!dir) {
    fprintf(stderr, "Error opening directory\n");
    return -1;
  }

  // Create a list of the image files
  struct dirent *ent;
  int num_frames = 0;
  char **frame_files = NULL;
  while ((ent = readdir(dir)) != NULL) {
    if (ent->d_type == DT_REG) {  // regular file
      num_frames++;
      frame_files = realloc(frame_files, num_frames * sizeof(char*));
      frame_files[num_frames-1] = strdup(ent->d_name);
    }
  }
  closedir(dir);

  // Sort the list of image files by filename
  qsort(frame_files, num_frames, sizeof(char*), compare_strings);










    
    /* find the mpeg1video encoder */
    //codec = avcodec_find_encoder_by_name(codec_name);
    codec = avcodec_find_encoder(AV_CODEC_ID_H264);
    if (!codec) {
        fprintf(stderr, "Codec '%s' not found\n", codec_name);
        exit(1);
    }

    c = avcodec_alloc_context3(codec);
    if (!c) {
        fprintf(stderr, "Could not allocate video codec context\n");
        exit(1);
    }

    pkt = av_packet_alloc();
    if (!pkt)
        exit(1);

    /* put sample parameters */
    c->bit_rate = 400000;
    /* resolution must be a multiple of two */
    c->width = 960;
    c->height = 540;
    /* frames per second */
    c->time_base = (AVRational){1, 25};
    c->framerate = (AVRational){25, 1};

    /* emit one intra frame every ten frames
     * check frame pict_type before passing frame
     * to encoder, if frame->pict_type is AV_PICTURE_TYPE_I
     * then gop_size is ignored and the output of encoder
     * will always be I frame irrespective to gop_size
     */
    c->gop_size = 10;
    c->max_b_frames = 1;
    c->pix_fmt = AV_PIX_FMT_YUV420P;

    if (codec->id == AV_CODEC_ID_H264)
        av_opt_set(c->priv_data, "preset", "slow", 0);

    /* open it */
    ret = avcodec_open2(c, codec, NULL);
    if (ret < 0) {
        fprintf(stderr, "Could not open codec: %s\n", av_err2str(ret));
        exit(1);
    }

    f = fopen(filename, "wb");
    if (!f) {
        fprintf(stderr, "Could not open %s\n", filename);
        exit(1);
    }

    frame = av_frame_alloc();
    if (!frame) {
        fprintf(stderr, "Could not allocate video frame\n");
        exit(1);
    }
    frame->format = c->pix_fmt;
    frame->width  = c->width;
    frame->height = c->height;

    ret = av_frame_get_buffer(frame, 0);
    if (ret < 0) {
        fprintf(stderr, "Could not allocate the video frame data\n");
        exit(1);
    }


    // new code



    if (av_frame_get_buffer(frame, 32) < 0) {
      fprintf(stderr, "Error allocating frame buffer\n");
      return -1;
    }

    













    
    // jpg decoding


    // Create an AVFrame to store the decoded JPEG image
    AVFrame *jpeg_frame = av_frame_alloc();
    if (!jpeg_frame) {
      fprintf(stderr, "Error allocating JPEG frame\n");
      return -1;
    }
    
    // Create an AVPacket to hold the compressed JPEG data
    AVPacket *jpg_packet = av_packet_alloc();
    if (!jpg_packet)
        exit(1);


    // Create a decoder for JPEG images
    const AVCodec *jpeg_codec = avcodec_find_decoder(AV_CODEC_ID_MJPEG);
    if (!jpeg_codec) {
      fprintf(stderr, "Error finding MJPEG decoder\n");
      return -1;
    }
    AVCodecContext *jpeg_codec_ctx = avcodec_alloc_context3(jpeg_codec);
    if (!jpeg_codec_ctx) {
      fprintf(stderr, "Error allocating MJPEG codec context\n");
      return -1;
    }
    jpeg_codec_ctx->pix_fmt = AV_PIX_FMT_YUV420P;

    
    if (avcodec_open2(jpeg_codec_ctx, jpeg_codec, NULL) < 0) {
      fprintf(stderr, "Error opening MJPEG codec\n");
      return -1;
    }

    /* printf("fr format: %d\n", frame->format); */
    /* printf("fr width: %d\n", frame->width); */
    /* printf("fr height: %d\n", frame->height); */


    jpeg_frame->format = frame->format;
    jpeg_frame->width = frame->width;
    jpeg_frame->height = frame->height;
    
    /* printf("jpg fr format: %d\n", jpeg_frame->format); */
    /* printf("jpg fr format: %d\n", jpeg_codec_ctx->pix_fmt); */
    /* printf("jpg fr width: %d\n", jpeg_frame->width); */
    /* printf("jpg fr height: %d\n", jpeg_frame->height); */


    struct SwsContext *sws_ctx = sws_getContext(jpeg_frame->width, jpeg_frame->height, jpeg_codec_ctx->pix_fmt,
						frame->width, frame->height, frame->format,
						SWS_BILINEAR, NULL, NULL, NULL);
    if (!sws_ctx) {
      fprintf(stderr, "Error creating software scaling context\n");
      return -1;
    }



    /* printf("fr format: %d\n", frame->format); */
    /* printf("fr width: %d\n", frame->width); */
    /* printf("fr height: %d\n", frame->height); */


    /* printf("jpg fr format: %d\n", jpeg_frame->format); */
    /* printf("jpg fr width: %d\n", jpeg_frame->width); */
    /* printf("jpg fr height: %d\n", jpeg_frame->height); */
    
    

    /* encode 1 second of video */
    for (i = 0; i < 115; i++) {
        fflush(stdout);

	/* printf("%s\n", frame_files[i]); */

	size_t len_image;
	char filename[200];
	snprintf(filename, 200, "input/%s", frame_files[i]);
	unsigned char *image_data = read_file(filename, &len_image);


	jpg_packet->data = image_data;
	jpg_packet->size = len_image;

	//Decode the JPEG data into the AVFrame
	int ret = avcodec_send_packet(jpeg_codec_ctx, jpg_packet);
	if (ret < 0) {
	  fprintf(stderr, "Error sending JPEG packet for decoding\n");
	  return -1;
	}

	/* printf("fr format: %d\n", frame->format); */
	/* printf("fr width: %d\n", frame->width); */
	/* printf("fr height: %d\n", frame->height); */

	
        ret = av_frame_make_writable(frame);
        if (ret < 0)
            exit(1);
	
	ret = avcodec_receive_frame(jpeg_codec_ctx, jpeg_frame);
	if (ret < 0) {
	  fprintf(stderr, "Error receiving decoded JPEG frame\n");
	  fprintf(stderr, "error %d\n", ret);

	  char errbuf[AV_ERROR_MAX_STRING_SIZE];
	  av_make_error_string(errbuf, AV_ERROR_MAX_STRING_SIZE, ret);
	  fprintf(stderr, "Error receiving frame: %s\n", errbuf);
	  return -1;
	}

	/* puts("after"); */
	/* printf("fr format: %d\n", frame->format); */
	/* printf("fr width: %d\n", frame->width); */
	/* printf("fr height: %d\n", frame->height); */



	ret = sws_scale(sws_ctx, (const uint8_t * const *) jpeg_frame->data,
			jpeg_frame->linesize, 0, jpeg_frame->height,
			frame->data, frame->linesize);
	if (ret < 0) {
	  fprintf(stderr, "Error scaling image\n");
	  return -1;
	}


	
	/* int ret = av_image_fill_arrays(frame->data, frame->linesize, image_data, AV_PIX_FMT_YUV420P, frame->width, frame->height, 1); */
	/* if (ret < 0) { */
	/*   fprintf(stderr, "Error setting up frame buffer: %s\n", av_err2str(ret)); */
	/*   return -1; */
	/* } */

	
        /* Make sure the frame data is writable.
           On the first round, the frame is fresh from av_frame_get_buffer()
           and therefore we know it is writable.
           But on the next rounds, encode() will have called
           avcodec_send_frame(), and the codec may have kept a reference to
           the frame in its internal structures, that makes the frame
           unwritable.
           av_frame_make_writable() checks that and allocates a new buffer
           for the frame only if necessary.
         */
        /* ret = av_frame_make_writable(frame); */
        /* if (ret < 0) */
        /*     exit(1); */

        /* Prepare a dummy image.
           In real code, this is where you would have your own logic for
           filling the frame. FFmpeg does not care what you put in the
           frame.
         */
        /* Y */
        /* for (y = 0; y < c->height; y++) { */
        /*     for (x = 0; x < c->width; x++) { */
        /*         frame->data[0][y * frame->linesize[0] + x] = x + y + i * 3; */
        /*     } */
        /* } */

        /* /\* Cb and Cr *\/ */
        /* for (y = 0; y < c->height/2; y++) { */
        /*     for (x = 0; x < c->width/2; x++) { */
        /*         frame->data[1][y * frame->linesize[1] + x] = 128 + y + i * 2; */
        /*         frame->data[2][y * frame->linesize[2] + x] = 64 + x + i * 5; */
        /*     } */
        /* } */

        frame->pts = i;

        /* encode the image */
        encode(c, frame, pkt, f);
    }

    /* flush the encoder */
    encode(c, NULL, pkt, f);

    /* Add sequence end code to have a real MPEG file.
       It makes only sense because this tiny examples writes packets
       directly. This is called "elementary stream" and only works for some
       codecs. To create a valid file, you usually need to write packets
       into a proper file format or protocol; see mux.c.
     */
    if (codec->id == AV_CODEC_ID_MPEG1VIDEO || codec->id == AV_CODEC_ID_MPEG2VIDEO)
        fwrite(endcode, 1, sizeof(endcode), f);
    fclose(f);

    avcodec_free_context(&c);
    av_frame_free(&frame);
    av_packet_free(&pkt);

    return 0;
}
